package Apache::AxKit::Language::FOP;

use strict;
use Apache ();
use Apache::Constants ();
use Apache::AxKit::Language ();
use Apache::AxKit::Cache ();
use Digest::MD5 ();
use File::Spec ();
use File::Temp ();
use File::Path ();
use File::stat ();
use File::Copy ();
use Fcntl qw( :mode );
@Apache::AxKit::Language::FOP::ISA = qw( Apache::AxKit::Language );

$Apache::AxKit::Language::FOP::VERSION = '0.91';

my $content_type = {
		    pdf => 'application/pdf',
		    mif => 'application/vnd.mif',
		    pcl => 'application/vnd.hp-PCL',
		    ps => 'application/postscript',
		    txt => 'text/plain',
		    svg => 'image/svg+xml',
		    at => 'text/xml',
		   };

sub stylesheet_exists { 0 }

sub handler {
    my $class = shift;
    my( $r, $xml_provider, undef, $last_in_chain ) = @_;

    my $javadir = $r->dir_config( 'AxFOPJavaServerDirectory' );
    my $dir = $javadir || File::Spec->tmpdir();
    my $tempdir = File::Temp::tempdir( 'axkit-fop-XXXXXX', DIR => $dir )
      or die "AxKit FOP: could not create tempdir under $dir: $!";
    AxKit::Debug( 8, "AxKit FOP: tempdir='$tempdir'" );
    $r->register_cleanup( sub { File::Path::rmtree( $tempdir ) } );
    if ( $javadir ) {
	my $stat = File::stat::stat( $tempdir )
	  or die "AxKit FOP: could not stat $tempdir: $!";
	my $mode = $stat->mode | S_IRWXG;
	chmod( $mode, $tempdir )
	  or die "AxKit FOP: could not set permissions on $tempdir: $!";
    }

    # Get input for FOP and write it to a file.
    my $input;
    my $input_file = File::Spec->catfile( $tempdir, 'input.fo' );
    eval {
	if ( my $dom = $r->pnotes( 'dom_tree' ) ) {
	    delete $r->pnotes()->{'dom_tree'};
	    $input = $dom->toString;
	}
	elsif ( $input = $r->pnotes( 'xml_string' ) ) {
	    delete $r->pnotes()->{'xml_string'};
	}
	else {
	    $input = ${ $xml_provider->get_strref() };
	}
    };
    if ( $@ ) {
	my $fh = $xml_provider->get_fh();
	File::Copy::copy( $fh, $input_file )
	    or die "AxKit FOP: could not write to $input_file: $!";
	close( $fh ) or die "AxKit FOP: could not close $input_file: $!";
    }
    else {
	my $fh = Apache->gensym();
	open( $fh, ">$input_file" )
	  or die "AxKit FOP: could not open $input_file: $!";
	print $fh $input or die "AxKit FOP: could not write to $input_file: $!";
	close( $fh ) or die "AxKit FOP: could not close $input_file: $!";
    }

    # Determine which output style is preferred.
    my $style = $AxKit::Cfg->PreferredStyle;
    unless ( defined $content_type->{$style} ) {
	AxKit::Debug( 1, "AxKit FOP: style '$style' unrecognized; using 'pdf'" );
	$style = 'pdf';
    }
    my $output_file = File::Spec->catfile( $tempdir, "output.$style" );

    # Determine if we should cache separately or leave AxKit to do it.
    my $axfopcache = lc $r->dir_config( 'AxFOPCache' ) || '';
    if ( $axfopcache eq 'ifnocache' ) {
	$axfopcache = $AxKit::Cache->no_cache() ? 1 : 0;
    }
    elsif ( $axfopcache eq 'always' ) {
	$axfopcache = 1;
    }
    else {
	$axfopcache = 0;
    }
    AxKit::Debug( 8, "AxKit FOP: AxFOPCache = $axfopcache" );

    # Create a cache object if caching is requested.
    my $cache;
    if ( $axfopcache ) {
	my $digest = Digest::MD5->new;
	my $fh = Apache->gensym();
	open( $fh, "<$input_file" )
	  or die "AxKit FOP: could not open $input_file: $!";
	$digest->addfile( $fh );
	close( $fh ) or die "AxKit FOP: could not close $input_file: $!";
	$digest->add( $style );
	my $extra = $digest->hexdigest;
	$cache = Apache::AxKit::Cache->new( $r, $xml_provider->key(), $extra );
    }

    # Check if output is in the cached (mtime is undef when not cached).
    my $fh;
    if ( $cache and $cache->mtime ) {
	AxKit::Debug( 1, "AxKit FOP: delivering output from cache for " .
		      $xml_provider->key() );
	$fh = $cache->get_fh;
    }
    else {
	run_fop( $r, $style, $input_file, $output_file );
	$fh = Apache->gensym();
	open( $fh, "<$output_file" )
	  or die "AxKit FOP: could not open $output_file: $!";
	if ( $cache ) {
	    # Cache the output for future requests.
	    AxKit::Debug( 1, "AxKit FOP: saving output in cache for " .
			  $xml_provider->key() );
	    $cache->write( do { local $/; <$fh> } );
	    seek( $fh, 0, 0 );
	}
    }

    # Read output and send it.
    $AxKit::Cfg->AllowOutputCharset(0);
    $r->content_type( $content_type->{$style} );
    if ( $last_in_chain ) {
	$r->print( do { local $/; <$fh> } );
    }
    else {
	$r->pnotes( 'xml_string' => do { local $/; <$fh> } );
    }
    close( $fh ) or die "AxKit FOP: could not close $output_file: $!";

    return Apache::Constants::OK;
}

#-------------------------------------------------------------------------------
sub run_fop {
    my( $r, $style, $input_file, $output_file ) = @_;

    my $config_file = $r->dir_config( 'AxFOPConfigFile' );
    my $javadir     = $r->dir_config( 'AxFOPJavaServerDirectory' );
    my $port        = $r->dir_config( 'AxFOPJavaServerPort' );
    my $authfile    = $r->dir_config( 'AxFOPJavaServerAuthFile' );
    if ( defined $javadir or defined $port or defined $authfile ) {
	require XML::ApacheFOP;
	my $debug = " port=$port" if defined $port;
	$debug .= " authfile=$authfile" if defined $authfile;
	AxKit::Debug( 8, "AxKit FOP: connecting to JavaServer$debug" );
	my $fop = XML::ApacheFOP->new( port => $port, authfile => $authfile );
	AxKit::Debug( 1, "AxKit FOP: running FOP via JavaServer: fo=$input_file, outfile=$output_file, rendertype=$style, c=$config_file" );
	$fop->fop( fo => $input_file, outfile => $output_file,
		   rendertype => $style, c => $config_file )
	  or die "AxKit FOP: could run FOP via JavaServer: " . $fop->errstr;
    }
    else {
	my $command = "JAVA_OPTS=-Djava.awt.headless=true" .
	  " fop -q $input_file -$style $output_file";
	$command .= " -c $config_file" if defined $config_file;
	AxKit::Debug( 1, "AxKit FOP: $command" );
	system( $command ) == 0 or die "AxKit FOP: fop died with status $?";
    }
}

1;

__END__

=head1 NAME

Apache::AxKit::Language::FOP - generate PDF using Apache FOP

=head1 SYNOPSIS

    AxAddStyleMap text/xslfo +Apache::AxKit::Language::FOP
    PerlSetVar AxFOPConfigFile /path/to/my/userconfig.xml

If you want it to use JavaServer, specify one or more of:

    PerlSetVar AxFOPJavaServerPort      2000
    PerlSetVar AxFOPJavaServerAuthFile  /path/to/my/authfile
    PerlSetVar AxFOPJavaServerDirectory /my/special/directory

If you want to cache when PDF is generated dynamically with XSP:

    PerlSetVar AxFOPCache IfNoCache

=head1 DESCRIPTION

This module uses Apache FOP to generate PDF by processing XSL-FO
files.  It can either call the command line tool "fop", or for better
performance, it can use the L<XML::ApacheFOP|XML::ApacheFOP> package
(which uses the L<Java.pm|Java> Perl package and JavaServer software).

To enable use of L<XML::ApacheFOP|XML::ApacheFOP>, specify one or more
AxFOPJavaServer directives as shown above.  Unspecified directives
will fall back to their default values defined in
L<XML::ApacheFOP|XML::ApacheFOP>.

If you need FOP to run with an external configuration file, specify
the location using the AxFOPConfigFile directive.  For example, you
can specify C<baseDir> in an external configuration file if you have a
C<fo:external-graphic> with a relative URL for the C<src> attribute.
See http://xmlgraphics.apache.org/fop/configuration.html for more
options you can specify in an external configuration file.

Multiple output render types are supported by FOP, to varying degrees:

    pdf: application/pdf
    mif: application/vnd.mif
    pcl: application/vnd.hp-PCL
    ps:  application/postscript
    txt: text/plain
    svg: image/svg+xml
    at:  text/xml

If the AxKit PreferredStyle matches one of the above styles, then that
style will be the render type output by this module.  If the
PreferredStyle does not match, then the output type will be pdf.

=head2 Caching

The standard AxKit caching mechanism automatically caches the output
in most cases (unless disabled with AxNoCache).  However, if your
content is being generated dynamically from an XSP script, then AxKit
will not cache the output.

To enable caching for XSP-generated content, set the configuration
directive AxFOPCache to "IfNoCache".  When this directive is set, the
output will be cached separately by this module whenever AxKit has
turned off its caching (that is, when $AxKit::Cache->no_cache()
returns a true value).  The cache file is managed using
L<Apache::AxKit::Cache|Apache::AxKit::Cache> with a key generated
using L<Digest::MD5|Digest::MD5> from the input content and the output
style.

To turn off separate caching by this module you can set S<"AxFOPCache
None"> (this is the default).

=head1 CAVEATS

A temporary directory with a prefix of "axkit-fop-" is created on each
request (using L<File::Temp|File::Temp>) to contain the input and
output files for FOP.  This directory is readable/writable only by the
web server process and it is automatically deleted in the mod_perl
cleanup phase at the end of each request.

If you are using XML::ApacheFOP and JavaServer, then the JavaServer
process needs read/write permission on this temporary directory.  One
way to allow this permission is to run your JavaServer process under
the same user as your web server.

An alternative method, if your JavaServer is running as a different
user from your web server, is to run your JavaServer under a group in
which the web server belongs, and create a special temporary directory
owned by this group with the group sticky bit set so that
subdirectories will also be owned by the group.  Inform this module of
that directory with this directive:

    PerlSetVar AxFOPJavaServerDirectory /my/special/directory

This module will turn on the group read/write permissions on temporary
subdirectories created under this special directory.

Here is an example setup using Debian GNU/Linux commands (if someone
can generalize this example let me know).  Assume your JavaServer
process is running under the user "javaserv" and the group "javaserv",
and your web server is running under the user "www-data" and the group
"www-data".

Add your web server user to the JavaServer group:

    adduser www-data javaserv

Create a temporary directory with the group sticky bit set:

    install -m 2770 -o javaserv -g javaserv -d /var/cache/javaserver

Update your Apache configuration with this directive:

    PerlSetVar AxFOPJavaServerDirectory /var/cache/javaserver

=head1 SEE ALSO

L<AxKit|AxKit>,
L<XML::ApacheFOP|XML::ApacheFOP>

=head1 AUTHOR

Ken Neighbors <ken@nsds.com>

Based on the Apache::AxKit::Language::PassiveTex module, but severely
rewritten.

=head1 COPYRIGHT and LICENSE

Copyright (c) 2005 Ken Neighbors.  All rights reserved.  This module
is free software; you can redistribute it and/or modify it under the
same terms as Perl itself.

=cut
